/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>

using namespace std;

#include "cx88_xml_config.h"
#include "cx88_file_capture.h"
#include "cx88_udp_capture.h"
#include "cx88_mpeg_driver_factory.h"

#define CX88_DEFAULT_UDP_PORT 8802
#define DEFAULT_XML_CONFIG "~/.cx88.xml"
#define DEFAULT_TUNER_CONFIG "~/.tuner_config"

extern char *optarg;

typedef struct
{
   pthread_mutex_t lock;
   pthread_cond_t signal;
} cx88_completion_arg;

typedef struct
{
   cx88_driver *driver;
   cx88_xml_config *conf;
   double threshold;
} cx88_scan_arg;

static int duration_capture(
   cx88_driver *driver,
   string &channel,
   unsigned int duration_s);
      
static int noninteractive_capture(
   cx88_driver *driver,
   string &channel,
   cx88_capture_params &params);

static int interactive_capture(
   cx88_driver *driver,
   string &channel,
   cx88_xml_config *xmlconf,
   unsigned short port,
   bool auto_start);

static void print_channel(void *arg, int error);
   
static void capture_complete(void *arg, int error);

static int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params);

int main(int argc, char **argv)
{
   int error = 0, optch = -1;
   char *device = NULL;
   cx88_driver *driver = NULL;
   cx88_xml_config *xmlconf = NULL;
   tuner_config tunerconf;
   capvec captures;
   string channel = CX88_XML_CHANNEL_DEFAULT;
   unsigned int duration_s = 0;
   size_t file_cache = CX88_DEFAULT_FILE_CACHE;
   size_t udp_mtu = CX88_DEFAULT_UDP_MTU;
   unsigned short remote_port = 0;
   bool auto_start = true;
   static const char *help = 
      "Options:\n"
      "\t-d <path>: [Required] Specify device from which to capture\n"
      "\t-c <channel | profile:channel>: [Optional] Specify initial channel, e.g. '7' or 'ATSC:7'\n"
      "\t-u <host-ip | host-ip:port>: [Optional] Send capture over UDP to provided host\n"
      "\t-f <filename>: [Optional] Send capture to specified file\n"
      "\t-m <number>: [Optional] Set UDP MTU size (in bytes).  Affects all subsequent -u options until next -m.  Default is 1500 bytes.\n"
      "\t-b <number>: [Optional] Set file capture cache size (in bytes).  Affects all subsequent -f options until next -b.  Default is 4MB.\n"
      "\t-x <filename>: [Optional] Specify XML configuration file; default is "DEFAULT_XML_CONFIG"\n"
      "\t-t <filename>: [Optional] Specify libtuner configuration file; default is "DEFAULT_TUNER_CONFIG"\n"
      "\t-n <value>: [Optional] Specify non-interactive capture for provided number of minutes\n"
      "\t-p <value>: [Optional] Specify port on which to listen for remote control connections\n"
      "\t-h: [Optional] Do not automatically start capturing (interactive captures only)\n"
      "\t-? : Display this message\n";
   while (!error && ((optch = getopt(argc, argv, "d:c:u:f:m:b:x:t:n:p:h?")) != -1))
   {
      switch (optch) 
      {
         case 'd':
            if (device != NULL)
            {
               CX88ERR << "Warning: ignoring duplicate device directive" << endl;
            }
            else if ((device = strdup(optarg)) == NULL)
            {
               CX88ERR << "Unable to store device filename." << endl;
               error = ENOMEM;
            }
            break;
         case 'c':
            try
            {
               channel = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store channel name.";
               error = ENOMEM;
            }
            break;
         case 'u':
         {
            char *hostname = strdup(optarg);
            unsigned short portnum = CX88_DEFAULT_UDP_PORT;
            if (hostname == NULL)
            {
               CX88ERR << "Unable to store UDP target address." << endl;
               error = ENOMEM;
               break;
            }      
            char *port = strchr(hostname, ':');
            if (port != NULL)
            {
               port[0] = '\0';
               portnum = (unsigned short) (strtoul(port + 1, NULL, 10));
            }
            cx88_udp_capture *cap = new cx88_udp_capture(hostname, portnum, error, udp_mtu);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store UDP capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store UDP capture object." << endl;
               error = ENOMEM;
            }
            free(hostname);
            break;
         }
         case 'f':
         {
            cx88_file_capture *cap = new cx88_file_capture(optarg, error, file_cache);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 'm':
            udp_mtu = strtoul(optarg, NULL, 10);
            break;
         case 'b':
            file_cache = strtoul(optarg, NULL, 10);
            break;
         case 'x':
            if (xmlconf != NULL)
            {
               CX88ERR << "Warning: ignoring duplicate cx88 configuration file" << endl;
               break;
            }
            xmlconf = new cx88_xml_config(optarg, tunerconf, error);
            if (xmlconf == NULL)
            {
               CX88ERR << "Unable to store XML configuration object." << endl;
               error = ENOMEM;
            }
            if (error)
            {
               delete xmlconf;
               xmlconf = NULL;
            }
            break;
         case 't':
            error = tunerconf.load_file(optarg);
            break;
         case 'n':
            duration_s = (unsigned long) (strtod(optarg, NULL) * 60);
            break;
         case 'p':
            remote_port = (unsigned short) (strtoul(optarg, NULL, 10));
            break;
         case 'h':
            auto_start = false;
            break;
         case '?':
            cout << help;
            break;
         default:
            break;
      }
   }
   if (!error)
   {
      if (device == NULL)
      {
         CX88ERR << "Error: You must specify a device (-d)." << endl;
         error = EINVAL;
      }
      if (captures.size() == 0)
      {
         CX88ERR << "Error: You must specify at least one capture (-u, -f)." << endl;
         error = EINVAL;
      }
   }
   if (!error && (xmlconf == NULL))
   {
      xmlconf = new cx88_xml_config(DEFAULT_XML_CONFIG, tunerconf, error);
      if (xmlconf == NULL)
      {  
         error = ENOMEM;  
      }
      if (error)
      {
         CX88ERR << "Unable to load cx88 XML configuration file" << endl;
      }
   }
   if (!error)
   {
      if (strstr(device, "cx88mpeg") != NULL)
      {
         driver = cx88_mpeg_driver_factory::create(
            *xmlconf,
            tunerconf,
            device,
            captures,
            error);
         if (!error && (driver == NULL))
         {
            CX88ERR << "Unable to create cx88 MPEG driver" << endl;
            error = ENOMEM;  
         }
      }
      else
      {
         CX88ERR << "Invalid device filename" << endl;
         error = EINVAL;
      }
   }
   if (!error)
   {
      if (duration_s == 0)
      {
         error = interactive_capture(driver, channel, xmlconf, remote_port, auto_start);
      }
      else
      {
         error = duration_capture(driver, channel, duration_s);  
      }
   }
   for (capvec::size_type i = 0; i < captures.size(); ++i)
   {
      delete captures[i];  
   }
   delete driver;
   free(device);
   delete xmlconf;
   return error;
}
static void capture_complete(void *arg, int error)
{
   cx88_completion_arg *completion = reinterpret_cast <cx88_completion_arg*> (arg);
   pthread_mutex_lock(&(completion->lock));
   pthread_cond_signal(&(completion->signal));
   pthread_mutex_unlock(&(completion->lock));
}

static int duration_capture(
   cx88_driver *driver,
   string &channel,
   unsigned int duration_s)
{
   cx88_capture_params params;
   params.duration_s = duration_s;
   params.channel_event = NULL;
   params.channel_arg = NULL;
   return noninteractive_capture(driver, channel, params);
}

static int noninteractive_capture(
   cx88_driver *driver,
   string &channel,
   cx88_capture_params &params)
{
   int error = 0;
   cx88_completion_arg arg;
   params.completion = capture_complete;
   params.completion_arg = &arg;
   if ((error = pthread_mutex_init(&(arg.lock), NULL)))
   {
      CX88ERR << "Unable to initialize non-interactive capture lock" << endl;
      return error;
   }
   if ((error = pthread_cond_init(&(arg.signal), NULL)))
   {
      pthread_mutex_destroy(&(arg.lock));
      CX88ERR << "Unable to initialize non-interactive capture signal" << endl;
      return error;
   }
   pthread_mutex_lock(&(arg.lock));
   if ((error = driver->start(channel, params)))
   {
      CX88ERR << "Unable to start driver" << endl;
      pthread_mutex_unlock(&(arg.lock));
      pthread_cond_destroy(&(arg.signal));
      pthread_mutex_destroy(&(arg.lock));
      return error;
   }
   pthread_cond_wait(&(arg.signal), &(arg.lock));
   pthread_mutex_unlock(&(arg.lock));
   driver->stop();
   pthread_cond_destroy(&(arg.signal));
   pthread_mutex_destroy(&(arg.lock));
   return error;
}

static void print_channel(void *arg, int error)
{
   cx88_scan_arg * scan_arg = reinterpret_cast<cx88_scan_arg *>(arg);
   usleep(500000);
   double strength = (error ? 0.0 : scan_arg->driver->get_signal_strength());
   if (strength >= scan_arg->threshold)
   {
      cout << scan_arg->conf->get_channel() << " (" << strength << "%) ";  
   }
   else
   {
      cout << ". ";  
   }
   if (scan_arg->conf->last_channel())
   {
      scan_arg->driver->stop();
   }
   else
   {
      string channel = CX88_XML_CHANNEL_UP;
      if (scan_arg->driver->set_channel(channel) != 0)
      {
         scan_arg->driver->stop();
      }
   }
}

static int interactive_capture(
   cx88_driver *driver,
   string &channel,
   cx88_xml_config *xmlconf,
   unsigned short remote_port,
   bool auto_start)
{
   int error = 0;
   string str = channel;
   cx88_capture_params params;
   params.completion = NULL;
   params.completion_arg = NULL;
   params.channel_event = NULL;
   params.channel_arg = NULL;
   params.duration_s = 0;
   fd_set fds;
   int server_sock = -1, client_sock = -1;
   bool halted = !auto_start;
   bool remote_input = false;
   bool started = false;
   int maxfd = fileno(stdin) + 1;
   if (remote_port != 0)
   {
      struct sockaddr_in server_addr;
      memset(&server_addr, 0, sizeof(server_addr));
      server_addr.sin_family = AF_INET;
      server_addr.sin_port = htons(remote_port);
      server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
      if ((server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
      {
         CX88ERR << "Unable to create remote control socket: " << strerror(errno) << endl;
         return errno;
      }
      if (bind(server_sock, (struct sockaddr*) (&server_addr), sizeof(server_addr)) < 0) 
      {
         close(server_sock);
         CX88ERR << "Unable to bind remote control socket: " << strerror(errno) << endl;
         return errno;
      }
      if (listen(server_sock, 1) < 0)
      {
         close(server_sock);
         CX88ERR << "Unable to establish socket listener: " << strerror(errno) << endl;
         return errno;
      }
      if (server_sock >= maxfd)
      {
         maxfd = server_sock + 1;
      }
   }
   if (!halted)
   {
      if ((error = driver->start(channel, params)))
      {
         CX88ERR << "Unable to start driver" << endl;
         return error;
      }
      started = true;
   }
   int cmd, nfds;
   do
   {
      cmd = 0;
      FD_ZERO(&fds);
      FD_SET(fileno(stdin), &fds);
      if (server_sock != -1)
      {
         FD_SET(server_sock, &fds);
         if (client_sock != -1)
         {
            FD_SET(client_sock, &fds);
         }
      }
      remote_input = false;
      if ((nfds = select(maxfd, &fds, NULL, NULL, NULL)) > 0)
      {
         if (FD_ISSET(fileno(stdin), &fds))
         {
            cmd = getchar();
         }
         else if ((server_sock != -1) && (FD_ISSET(server_sock, &fds)))
         {
            struct sockaddr_in client_addr;
            socklen_t client_addr_len;
            if (client_sock != -1)
            {
               close(client_sock);
            }
            client_sock = accept(server_sock, (struct sockaddr *)(&client_addr), &client_addr_len);
            if (client_sock >= maxfd)
            {
               maxfd = client_sock + 1;
            }
            continue;
         }
         else if ((client_sock != -1) && (FD_ISSET(client_sock, &fds)))
         {
            char data = '\0';
            ssize_t count = recv(client_sock, &data, sizeof(data), 0);
            if (count <= 0)
            {
               continue;  
            }
            cmd = data;
            remote_input = true;
         }
      }
      if (remote_input)
      {
         switch (cmd)
         {
            case 'c':
            case 's':
            {
               char c;
               ssize_t count = 0;
               str = "";
               do
               {
                  if ((count = recv(client_sock, &c, sizeof(c), 0)) == 1)
                  {
                     if (c != '\0')
                     {
                        str += c;
                     }
                  }
               } while ((count == 0) || (c != '\0'));
               break;
            }
         }
      }
      else
      {
         switch (cmd)
         {
            case 'c':  
               cout << "Enter channel:" << endl;
               getline(cin, str);
               getline(cin, str);
               break;
            case 's':
               cout << "Enter signal strength threshold:" << endl;
               getline(cin, str);
               getline(cin, str);
               break;
         }
      }
      switch (cmd)
      {
         case 'c':
            if (driver->set_channel(str) != 0)
            {
               CX88ERR << "Unable to change channel" << endl;
            }
            break;
         case 'i':
            if (driver->get_info(str) != 0)
            {
               CX88ERR << "Unable to get driver info" << endl;  
            }
            else
            {
               cout << str.c_str() << endl;
            }
            break;
         case 's':
         {
            if (halted)
            {
               CX88ERR << "Unable to scan: driver is halted" << endl;
               break;  
            }
            stringstream stream(str);
            cx88_scan_arg arg;
            arg.driver = driver;
            arg.conf = xmlconf;
            stream >> arg.threshold;
            scan(arg, params);
            break;  
         }
         case 'h':
            if (!halted)
            {
               driver->stop();  
               halted = true;
            }
            else
            {
               if (!started)
               {
                  str = channel;
               }
               else
               {
                  xmlconf->get_channel_name(str);
               }
               if ((error = driver->start(str, params)))
               {
                  CX88ERR << "Unable to restart driver" << endl;
               }
               halted = false;
               started = true;
            }
            break;
         default:
            break;
      }
   } while (!error && (cmd != 'q'));
   driver->stop();
   if (client_sock != -1)
   {
      close(client_sock);
   }
   if (server_sock != -1)
   {
      close(server_sock);
   }
   return error;
}

int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params)
{
   int error = 0;
   string orig_channel;
   arg.conf->get_channel_name(orig_channel);
   arg.driver->stop();
   cx88_capture_params scan_params;
   scan_params.channel_event = print_channel;
   scan_params.channel_arg = &arg;
   scan_params.duration_s = 0;
   cout << "Scanning profile " << arg.conf->get_profile_name() << ' ';
   string start_channel = CX88_XML_CHANNEL_FIRST;
   ofstream temperr("/dev/null");
   streambuf *origerr = cerr.rdbuf(temperr.rdbuf());
   noninteractive_capture(arg.driver, start_channel, scan_params);
   cout << "Finished scan." << endl;
   cerr.rdbuf(origerr);
   if ((error = arg.driver->start(orig_channel, orig_params)))
   {
      CX88ERR << "Unable to restart driver" << endl;
      return error;
   }
   return error;
}
