/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/endian.h>
#include "cx88_risc.h"

/* RISC opcodes */
typedef enum
{
   CX88_RISC_WRITE = 0x1,
   CX88_RISC_SKIP = 0x2,
   CX88_RISC_WRITEC = 0x5,
   CX88_RISC_JUMP = 0x7,
   CX88_RISC_SYNC = 0x8,
   CX88_RISC_READ = 0x9,
   CX88_RISC_READC = 0xA,
   CX88_RISC_WRITERM = 0xB,
   CX88_RISC_WRITECM = 0xC,
   CX88_RISC_WRITECR = 0xD
} cx88_risc_opcode;

/* Line position indicator */
enum
{
   CX88_RISC_EOL = 0x1, /* End of line */
   CX88_RISC_SOL /* Start of line */
};

static uint8_t cx88_get_linepos(
   cx88_buffer *buffer,
   uint32_t *count);
   
static void cx88_write_riscinstr(cx88_buffer *buffer, uint32_t instr);

/* Generate a RISC instruction */
static inline uint32_t cx88_format_instr(
   cx88_risc_opcode opcode,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   return ((opcode << 28) | (intrs << 24) | (countctl << 16));
}

/* Write a RISC instruction to memory */
static void cx88_write_riscinstr(cx88_buffer *buffer, uint32_t instr)
{
   if (buffer->dma_program_buffer == NULL)
   {
      /* There's no RAM buffer, so we should place the instruction in SRAM */
      cx88_write_reg(
         buffer->channel->sc,
         buffer->dma_program_physaddr + (buffer->dma_program_counter * sizeof(uint32_t)), 
         instr);
      ++(buffer->dma_program_counter);
   }
   else
   {
      /* Place the instruction in RAM, in PCI little-endian format */
      buffer->dma_program_buffer[buffer->dma_program_counter++] = htole32(instr); 
   }
}

/* Determine the maximum number of bytes that can be skipped/transferred with a single
 * instruction, return that number of bytes (via count) and return the line position
 * (SOL/EOL) flags required by the instruction. */
static uint8_t cx88_get_linepos(
   cx88_buffer *buffer,
   uint32_t *count)
{
   uint8_t linepos = 0;
   uint32_t actualcount = (*count);
   DIAGNOSTIC(printf("CX88-RISC: Getting line position for 0x%x-byte transfer\n", actualcount))
   if ((buffer->sgl_index->ds_len - buffer->sgl_offset) < actualcount)
   {
      /* A single instruction can only skip/transfer until the end of the buffer segment */
      actualcount = buffer->sgl_index->ds_len - buffer->sgl_offset;
   }
   if (buffer->channel->line_offset == 0)
   {
      DIAGNOSTIC(printf("CX88-RISC: Start of line\n"))
      linepos |= CX88_RISC_SOL;
   }
   if ((buffer->channel->line_offset + actualcount) >= buffer->channel->line_size)
   {
      DIAGNOSTIC(printf("CX88-RISC: End of line\n"))
      linepos |= CX88_RISC_EOL;
      /* A single instruction can also only skip/transfer until the end of the current line */
      actualcount = buffer->channel->line_size - buffer->channel->line_offset;
   }
   buffer->channel->line_offset += actualcount;
   if (buffer->channel->line_offset == buffer->channel->line_size)
   {
      /* If we've used up the current line, reset the line offset to indicate
       * beginning of the next line for the next skip/transfer */
      buffer->channel->line_offset = 0;
   }
   buffer->sgl_offset += actualcount;
   if (buffer->sgl_offset == buffer->sgl_index->ds_len)
   {
      /* If we've used up the current buffer segment, reset the segment offset
       * and move to the next segment for the next skip/transfer */
      buffer->sgl_offset = 0;
      ++(buffer->sgl_index);
   }
   buffer->dma_byte_counter += actualcount;
   *count = actualcount;
   DIAGNOSTIC(printf("CX88-RISC: Able to transfer 0x%x bytes\n", actualcount));
   return linepos;
}

/* Generic single-instruction transfer */
uint32_t cx88_risc_transfer(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   /* Determine whether to use READ(C) or WRITE(C) */
   boolean_t is_read = ((buffer->channel->descriptor->direction == CX88_UPSTREAM) ? TRUE : FALSE);
   uint32_t actualcount = count;
   bus_size_t sgl_offset = buffer->sgl_offset;
   /* Calculate target physical address and maximum transfer count */
   uint32_t physaddr = (uint32_t) (buffer->sgl_index->ds_addr + sgl_offset);
   uint8_t linepos = cx88_get_linepos(buffer, &actualcount);
   boolean_t is_continued = FALSE;
   if ((!(linepos & CX88_RISC_SOL)) && (sgl_offset > 0))
   {
      /* If we're not at the start of the line or the DMA segment, we
       * don't need to supply the physical address and can use WRITEC or READC */
      is_continued = TRUE;
   }
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      /* We're actually writing the DMA program, so generate the instruction */
      cx88_risc_opcode operation;
      if (is_continued)
      {
         operation = (is_read ? CX88_RISC_READC : CX88_RISC_WRITEC);
         DIAGNOSTIC(is_read ? printf("READC:") : printf("WRITEC:"))
      }
      else
      {
         operation = (is_read ? CX88_RISC_READ : CX88_RISC_WRITE);
         DIAGNOSTIC(is_read ? printf("READ:") : printf("WRITE:"))
      }
      DIAGNOSTIC(printf("\t0x%x bytes\n", actualcount))
      uint32_t instr = cx88_format_instr(operation, intrs, countctl); 
      instr |= ((linepos & 0x3) << 26);
      instr |= (actualcount & 0xFFF);
      cx88_write_riscinstr(buffer, instr);
      if (!is_continued)
      {
         cx88_write_riscinstr(buffer, physaddr);
      }
   }
   else
   {
      /* We're sizing the program, so just bump the counter */
      buffer->dma_program_counter += 
         (is_continued ? CX88_RISC_READC_WRITEC_SIZE : CX88_RISC_READ_WRITE_SIZE);
   }
   return actualcount;
}

/* Sync to a specified line number */
void cx88_risc_sync(
   cx88_buffer *buffer,
   uint16_t lineno,
   boolean_t resync,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   buffer->channel->line_offset = 0;
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      DIAGNOSTIC(printf("SYNC to line %d\n", lineno))
      uint32_t sync_instr = cx88_format_instr(CX88_RISC_SYNC, intrs, countctl);
      if (resync)
      {
         sync_instr |= (1 << 15);
      }
      sync_instr |= (lineno & 0x3FF);
      cx88_write_riscinstr(buffer, sync_instr);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_SYNC_SIZE;
   }
}

/* Jump to arbitrary address in destination buffer's RISC program */
int cx88_risc_jump(
   cx88_buffer *src,
   cx88_buffer *dest,
   uint32_t offset,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (src->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t target;
      if ((dest == NULL) || (dest->data_buffer_type == CX88_DATABUF_NONE))
      {
         return EINVAL;
      }
      /* Calculate physical target address */
      target = dest->dma_program_physaddr + (offset * sizeof(uint32_t));
      DIAGNOSTIC(printf("JUMP to 0x%x\n", target))
      uint32_t jump_instr = cx88_format_instr(CX88_RISC_JUMP, intrs, countctl);
      if (dest->dma_program_buffer == NULL)
      {
         /* Specify a jump into an SRAM-resident program */
         jump_instr |= 1;
      }
      cx88_write_riscinstr(src, jump_instr);
      cx88_write_riscinstr(src, target);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      src->dma_program_counter += CX88_RISC_JUMP_SIZE;
   }
   return 0;
}

/* Skip as many bytes as possible (up to count) with single instruction */
uint32_t cx88_risc_skip(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t actualcount = count;
   uint8_t linepos = cx88_get_linepos(buffer, &actualcount);
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      DIAGNOSTIC(printf("SKIP 0x%x bytes\n", count))
      uint32_t skip_instr = cx88_format_instr(CX88_RISC_SKIP, intrs, countctl);
      skip_instr |= ((linepos & 0x3) << 26);
      skip_instr |= (count & 0xFFF) ;
      cx88_write_riscinstr(buffer, skip_instr);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_SKIP_SIZE;
   }
   return actualcount;
}

/* Write cx88 register to host memory */
void cx88_risc_write_reg_to_mem(
   cx88_buffer *buffer,
   uint32_t reg,
   uint32_t target_addr,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writerm_instr = cx88_format_instr(CX88_RISC_WRITERM, intrs, countctl);
      writerm_instr |= 1;
      cx88_write_riscinstr(buffer, writerm_instr);
      cx88_write_riscinstr(buffer, target_addr);
      cx88_write_riscinstr(buffer, (reg & 0xFFFFFF));
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITERM_SIZE;
   }
}

/* Write constant to host memory */
void cx88_risc_write_const_to_mem(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t target_addr,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writecm_instr = cx88_format_instr(CX88_RISC_WRITECM, intrs, countctl);
      cx88_write_riscinstr(buffer, writecm_instr);
      cx88_write_riscinstr(buffer, target_addr);
      cx88_write_riscinstr(buffer, constant);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITECM_SIZE;
   }
}

/* Write constant to cx88 register */
void cx88_risc_write_const_to_reg(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t mask,
   uint32_t reg,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   if (buffer->dma_program_state == CX88_DMAPROG_LOAD)
   {
      uint32_t writecr_instr = cx88_format_instr(CX88_RISC_WRITECR, intrs, countctl);
      writecr_instr |= 1;
      cx88_write_riscinstr(buffer, writecr_instr);
      cx88_write_riscinstr(buffer, (reg & 0xFFFFFF));
      cx88_write_riscinstr(buffer, constant);
      cx88_write_riscinstr(buffer, mask);
   }
   else /* Just sizing the DMA program, not writing it */
   {
      buffer->dma_program_counter += CX88_RISC_WRITECR_SIZE;
   }
}

/* Transfer arbitrary number of bytes */
int cx88_risc_transfer_dmabuf(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t bytes_transferred = 0;
   if ((buffer->dma_byte_counter + count) > (buffer->data_buffer_size))
   {
      return EINVAL;
   }
   while (bytes_transferred < count)
   {
      uint32_t bytes_remaining = count - bytes_transferred;
      bytes_transferred += cx88_risc_transfer (buffer, bytes_remaining, intrs, countctl);
   }
   return 0;
}

/* Skip arbitrary number of bytes */
int cx88_risc_skip_dmabuf(
   cx88_buffer *buffer,
   uint32_t count,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   uint32_t bytes_skipped = 0;
   if ((buffer->dma_byte_counter + count) > (buffer->data_buffer_size))
   {
      return EINVAL;
   }
   while (bytes_skipped < count)
   {
      uint32_t bytes_remaining = count - bytes_skipped;
      bytes_skipped += cx88_risc_skip (buffer, bytes_remaining, intrs, countctl);
   }
   return 0;
}

/* Return the greatest number of packets less than or equal to num_packets
 * that can fit into a whole number of lines */
uint32_t cx88_risc_check_bufsize(cx88_channel *channel, uint32_t num_packets)
{
   if (channel->packet_size != channel->line_size)
   {
      return (num_packets - (num_packets % (channel->line_size / channel->packet_size)));
   }
   else
   {
      return num_packets;
   }
}
