/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include "cx88_common.h"
#include "cx88_risc.h"
#include "cx88_if.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_video_ioctl.h"

#define MAX_VIDEO_BUFFER_SIZE (1 << 20)

/* Driver states */
typedef enum
{
   CX88_VIDEO_INVALID = 0, /* Uninitialized */
   CX88_VIDEO_UNCONFIGURED, /* Channels/buffers/I2C devices not yet configured */
   CX88_VIDEO_CONFIGURED, /* Configured but not formatted */
   CX88_VIDEO_FORMATTED, /* Formatted but not capturing */
   CX88_VIDEO_CAPTURING /* Configured, formatted, and capturing */
} cx88_video_state_t;

typedef struct
{
   cx88_common_softc common;
   cx88_channel ch_packed;
   cx88_channel ch_u;
   cx88_channel ch_v;
   cx88_channel ch_vbi;
   struct cdev *devnode;
   struct mtx isrlock;
   boolean_t data_avail;
   struct sx sxlock;
   struct selinfo selectors;
   cx88_i2c_bus i2c;
   cx88_i2c_dev i2c_devices[CX88_VIDEO_MAX_I2C_DEVICES];
   uint8_t num_i2c_devices;
   boolean_t in_use;
   boolean_t reconfig_requested;
   boolean_t has_user_buffers;
   boolean_t is_persistent;
   cx88_video_state_t state;
} cx88_video_softc;

static int cx88_video_probe(device_t dev);
static int cx88_video_attach(device_t dev);
static int cx88_video_setup (void *arg);
static int cx88_video_teardown (void *arg);
static void cx88_video_intr (void *arg);
static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr);
static int cx88_video_request_reconfig(device_t dev);
static void cx88_video_release_devices(cx88_video_softc *sc);
static int cx88_video_request_devices(cx88_video_softc *sc);

static const cx88_client_driver cx88_video_client = 
{
   .softc_size               = sizeof(cx88_video_softc),
   .function                 = CX88_FUNC_VIDEO,
   .setup                    = cx88_video_setup,
   .teardown                 = cx88_video_teardown,
   .interrupt_handler        = cx88_video_intr,
   .description              = "Conexant CX2388x Analog Video"
};

static devclass_t cx88_video_devclass;

static d_open_t cx88_video_open;
static d_close_t cx88_video_close;
static d_mmap_t cx88_video_mmap;
static d_ioctl_t cx88_video_ioctl;
static d_poll_t cx88_video_poll;

static struct cdevsw cx88_video_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88video",
   .d_open = cx88_video_open,   /* open */
   .d_close = cx88_video_close, /* close */
   .d_mmap = cx88_video_mmap,   /* mmap */
   .d_ioctl = cx88_video_ioctl, /* ioctl */
   .d_poll = cx88_video_poll,   /* poll */
};

static int cx88_video_setup (void *arg)
{   
   cx88_video_softc *sc = arg;
   int error = 0;
   sc->is_persistent = FALSE;
   sc->reconfig_requested = FALSE;
   sc->has_user_buffers = FALSE;
   sc->in_use = FALSE;
   mtx_init(&(sc->isrlock), "CX88 Video ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Video SX lock");
   //int unit = device_get_unit(sc->common.dev);
   if ((error = cx88_init_channel(&(sc->ch_packed), &(sc->common), CX88_CH_VIDEO_PACKED, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail0;
   }
   if ((error = cx88_init_channel(&(sc->ch_u), &(sc->common), CX88_CH_VIDEO_U, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail1;
   }
   if ((error = cx88_init_channel(&(sc->ch_v), &(sc->common), CX88_CH_VIDEO_V, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail2;
   }
   if ((error = cx88_init_channel(&(sc->ch_vbi), &(sc->common), CX88_CH_VIDEO_VBI, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail3;
   }
   if ((error = cx88_i2c_create(&(sc->i2c), &(sc->common))))
   {
      goto fail4;
   }
   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_video_cdevsw, device_get_unit(sc->common.dev), UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(sc->common.dev));
   if (sc->devnode == NULL)
   {
      goto fail5;
   }
   goto success;
   fail5:
      cx88_i2c_destroy(&(sc->i2c));
   fail4:
      cx88_destroy_channel (&(sc->ch_vbi));
   fail3:
      cx88_destroy_channel (&(sc->ch_v));
   fail2:
      cx88_destroy_channel (&(sc->ch_u));
   fail1:
      cx88_destroy_channel (&(sc->ch_packed));
   fail0:
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   success:
      return error;
}

static int cx88_video_teardown (void *arg)
{
   int error = 0;
   cx88_video_softc *sc = arg;
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sc->devnode = NULL;  
   }
   error = cx88_i2c_destroy(&(sc->i2c));
   if (!error)
   {
      cx88_destroy_channel (&(sc->ch_vbi));
      cx88_destroy_channel (&(sc->ch_v));
      cx88_destroy_channel (&(sc->ch_u));
      cx88_destroy_channel (&(sc->ch_packed));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

/* Open file operation */
static int cx88_video_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   int error = 0;
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }
   if (sc->state >= CX88_VIDEO_CONFIGURED)
   {
      /* If this device has already been configured, go ahead and
       * request ownership of I2C devices */
      error = cx88_video_request_devices(sc);
   }
   if (error == 0)
   {
      sc->in_use = TRUE;  
   }
   sx_xunlock(&sc->sxlock);
   return error;
}

/* Close file operation */
static int cx88_video_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{   
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   /*cx88_video_reset(sc, sc->state);
   if (sc->has_user_buffers)
   {*/
      /* If we're using user-allocated buffers, release them as
       * they can't be persisted across open() and close() in
       * different processes */
   /*   cx88_release_buffers(&(sc->channel));
      sc->has_user_buffers = FALSE;
   }
   if (sc->reconfig_requested || (!sc->is_persistent))
   {*/
      /* If resource reallocation has been requested by a sibling, or
       * we just aren't allowed to persist the configuration, 
       * release all channel resources and remove I2C devices */
      /*cx88_video_unconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   else
   {*/
      /* We can persist our configuration, so just release ownership
       * of I2C devices */
      /*cx88_video_release_devices(sc);
   }*/
   sc->in_use = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Mmap file operation */
static int cx88_video_mmap(
   struct cdev *dev,
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   return 0;
}

/* Ioctl file operation */
static int cx88_video_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   return 0;
}

/* Select/poll file operation */
static int cx88_video_poll(struct cdev *dev, int events, struct thread *td)
{
   return 0;
}

/* Release ownership of I2C devices */
static void cx88_video_release_devices(cx88_video_softc *sc)
{
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      cx88_i2c_release(&(sc->i2c_devices[i]));
   }
}

/* Request ownership of I2C devices */
static int cx88_video_request_devices(cx88_video_softc *sc)
{
   int error = 0;
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      error = cx88_i2c_request(&(sc->i2c_devices[i]));
      if (error)
      {
         /* If we can't own all of the devices, then we shouldn't own any */
         cx88_video_release_devices(sc);
         return error;
      }
   }
   return error;
}

static void cx88_video_intr (void *arg)
{
}

static int cx88_video_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx88_video_client);
}

static int cx88_video_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_video_client);
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return (&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr)
{
   uint8_t i;
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      /* Determine whether we have a device with the given address */
      if (sc->i2c_devices[i].addr == addr)
      {
         return (&(sc->i2c_devices[i]));
      }  
   }
   return NULL;
}

static int cx88_video_request_reconfig(device_t dev)
{
   return ENODEV;   
}

static device_method_t cx88_video_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_video_probe),
   DEVMETHOD(device_attach,            cx88_video_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_common_detach),
   DEVMETHOD(device_suspend,           cx88_common_detach),
   DEVMETHOD(device_resume,            cx88_video_attach),

   /* cx88 interface */
   CX88_METHODS(cx88_video_request_reconfig),

   /* i2c host methods */
   CX88_I2C_METHODS(cx88_video_get_i2c_bus, cx88_video_get_i2c_dev),

   {NULL, NULL}
};

static driver_t cx88_video_driver =
{
   "cx88video",
   cx88_video_methods,
   0
};

DRIVER_MODULE(cx88video, pci, cx88_video_driver, cx88_video_devclass, NULL, NULL);
DRIVER_MODULE(cx88video, cardbus, cx88_video_driver, cx88_video_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88video);
CX88_I2C_DRIVER_MODULE(cx88video);
MODULE_VERSION(cx88video, 1);
